// -*- Mode: C++; indent-tabs-mode: nil; tab-width: 2 -*-
/*
 * Copyright (C) 2014 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Brandon Schaefer <brandontschaefer@gmail.com>
 */

#include "Colors.h"
#include "SDLHandler.h"
#include "BlocksGrid.h"

#include "PieceFactory.h"

#include <algorithm>
#include <iostream>
#include <stdlib.h>

#include <unistd.h>

using namespace std;

namespace blocks
{

float const ROW_CLEAR_DURATION = 0.5f;
float const ROW_CLEAR_SPEED    = 0.09f;
int const FADDING_SIZE         = 5;
Color const FADDING_COLORS[]   = {LIGHT_GRAY, SILVER, GRAY, DARK_GRAY, BLACK};

BlocksGrid::BlocksGrid(Point const& start, Point const& xy_cells, int cell_size, int spacing)
  : Grid(start, xy_cells, cell_size, spacing)
  , score_(0)
  , lost_(false)
  , fadding_index_(0)
  , row_clearing_timer_(ROW_CLEAR_DURATION, ROW_CLEAR_SPEED)
{
  row_clearing_timer_.update.connect(sigc::mem_fun(this, &BlocksGrid::RowClearingUpdate));
  row_clearing_timer_.done.connect  (sigc::mem_fun(this, &BlocksGrid::RowClearingDone));
}

void BlocksGrid::Lose()
{
  if (!lost_)
  {
    printf("[BlocksGrid] You Lose, End Score: %i\n", score_);
    lost_ = true;

    Color color = GetRandomColor();
    for (auto piece : pieces_)
      for (auto p : piece->GetPoints())
        SetColor(p, color);
  }
}

void BlocksGrid::Reset()
{
  for (auto& piece : pieces_)
    ClearPieceRegion(piece);

  pieces_.clear();
  rows_to_clear_.clear();
  row_clearing_timer_.Stop();
  lost_ = false;
}

bool BlocksGrid::Lost()
{
  return lost_;
}

void BlocksGrid::AddNewPiece(Piece::Ptr const& piece)
{
  pieces_.push_back(piece);
}

void BlocksGrid::RemovePiece(Piece::Ptr const& piece)
{
  pieces_.erase(std::remove(pieces_.begin(), pieces_.end(), piece), pieces_.end());
}

bool BlocksGrid::CanPieceRotateClockwise(Piece::Ptr& piece) const
{
  if (!piece->CanRotate())
    return false;

  piece->RotateClockwise();
  auto points = piece->GetPoints();
  piece->UndoRotation();

  for (auto p : points)
    if (IsMarked(p) && !piece->PointInPiece(p))
      return false;

  return true;
}

bool BlocksGrid::CanPieceMoveDirection(Piece::Ptr& piece, Direction direction) const
{
  piece->MoveDirection(direction);
  auto points = piece->GetPoints();
  piece->MoveOppositeDirection(direction);

  for (auto p : points)
    if (IsMarked(p) && !piece->PointInPiece(p))
      return false;

  return true;
}

bool BlocksGrid::Animating() const
{
  return row_clearing_timer_.IsRunning();
}

void BlocksGrid::Update(float delta_time)
{
  if (row_clearing_timer_.IsRunning())
    row_clearing_timer_.Tick(delta_time);
}

void BlocksGrid::ClearPieceRegion(Piece::Ptr const& piece)
{
  for (auto& p : piece->GetPoints())
  {
    SetPhased(p);
    SetColor(p, BLACK);
  }
}

void BlocksGrid::UpdatePieceRegion(Piece::Ptr const& piece)
{
  for (auto& p : piece->GetPoints())
  {
    if (!piece->IsGhostPiece())
      UnsetPhased(p);

    SetColor(p, piece->GetColor());
  }
}

bool BlocksGrid::RowsNeedClearing()
{
  if (!rows_to_clear_.empty())
    return true;

  bool entire_row_marked = false;

  for (int y = 1; y < GetYCells() - 1; ++y)
  {
    entire_row_marked = true;
    for (int x = 1; x < GetXCells(); ++x)
    {
      if (!IsMarked({x, y}))
      {
        entire_row_marked = false;
        break;
      }
    }

    if (entire_row_marked)
    {
      rows_to_clear_.push_back(y);
    }
  }

  bool need_clearing = !rows_to_clear_.empty();

  if (need_clearing)
  {
    row_clearing_timer_.Start();
  }

  return need_clearing;
}

void BlocksGrid::RowClearingUpdate()
{
  Color const& color = FADDING_COLORS[fadding_index_ % FADDING_SIZE];

  for (auto const& row : rows_to_clear_)
  {
    for (auto const& piece : pieces_)
    {
      for (auto const& p : piece->GetPoints())
      {
        if (p.y() == row)
        {
          SetColor(p, color);
        }
      }
    }
  }

  fadding_index_++;
}

void BlocksGrid::RowClearingDone()
{
  for (int const& row : rows_to_clear_)
  {
    for (auto piece : pieces_)
    {
      for (auto const& p : piece->GetPoints())
      {
        if (p.y() == row)
        {
          ClearPieceRegion(piece);
          piece->RemovePoint(p);
          UpdatePieceRegion(piece);
        }
      }
    }
  }

  for (int row = (int)rows_to_clear_.size() - 1; row >= 0; --row)
  {
    for (auto piece : pieces_)
    {
      ClearPieceRegion(piece);
      piece->MoveDownPointsAboveRow(rows_to_clear_[row]);
      UpdatePieceRegion(piece);
    }
  }

  score_ += (unsigned)rows_to_clear_.size() * (unsigned)rows_to_clear_.size() * 10;
  printf("[BlocksGrid] Current Score: %i\n", score_);

  fadding_index_ = 0;
  rows_to_clear_.clear();
  animation_done.emit();
}

void BlocksGrid::Draw(SDLHandler* sdl_handler)
{
  Grid::Draw(sdl_handler);
}

}
