// -*- Mode: C++; indent-tabs-mode: nil; tab-width: 2 -*-
/*
 * Copyright (C) 2014 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Brandon Schaefer <brandontschaefer@gmail.com>
 */

#include "BlocksGridController.h"
#include "Colors.h"
#include "SDLHandler.h"
#include "IPiece.h"

#include <limits>

namespace blocks
{

int const GRID_X = 2;
int const GRID_Y = 0;

int const NGRID_X = 5;
int const NGRID_Y = 22;

// FIXME We had to increase the oversize by 1...so we get 20 squares
Point const START      = {4,  22};
Point const NP_START   = {1,   3};
Point const GRID_SIZE  = {12, 23};
Point const NGRID_SIZE = {6,   5};

BlocksGridController::BlocksGridController(ScreenGrid const& screen, int width, int height)
  : width_(width)
  , height_(height)
  , screen_(screen)
  , center_x_(screen_.GetCenterX(GRID_SIZE.x()))
  , start_   (screen_.GetX(GRID_X)  + center_x_,  screen_.GetY(GRID_Y))
  , nstart_  (screen_.GetX(NGRID_X) + center_x_, screen_.GetY(NGRID_Y))
  , grid_           (start_,  GRID_SIZE,  screen_.CellSize(), 1)
  , next_piece_grid_(nstart_, NGRID_SIZE, screen_.CellSize(), 1)
  , current_piece_seed_(piece_factory_.GetRandomSeed())
  , next_piece_seed_   (piece_factory_.GetRandomSeed())
  , current_piece_     (piece_factory_.GenerateRandomPiece(START, current_piece_seed_))
  , ghost_piece_       (current_piece_)
  , next_piece_        (piece_factory_.GenerateRandomPiece(NP_START, next_piece_seed_))
  , can_swap_(true)
  , ghost_piece_active_(true)
  , dir_(DOWN)
{
  grid_.AddNewPiece(current_piece_);
  next_piece_grid_.AddNewPiece(next_piece_);

  next_piece_grid_.ClearPieceRegion(next_piece_);
  next_piece_grid_.UpdatePieceRegion(next_piece_);

  UpdateGhostPiece();

  grid_.animation_done.connect(sigc::mem_fun(this, (&BlocksGridController::PickNewCurrentPiece)));
}

void BlocksGridController::DeepCopyCurrentPieceToGhostPiece()
{
  ghost_piece_ = piece_factory_.GenerateRandomPiece(START, current_piece_seed_);
  *ghost_piece_ = *current_piece_.get();
  ghost_piece_->SetAsGhostPiece();
}

void BlocksGridController::UpdateGhostPiece()
{
  if (ghost_piece_active_)
  {
    grid_.ClearPieceRegion(ghost_piece_);

    DeepCopyCurrentPieceToGhostPiece();

    while (grid_.CanPieceMoveDirection(ghost_piece_, DOWN) && !grid_.Lost())
      ghost_piece_->MoveDirection(DOWN);

    grid_.UpdatePieceRegion(ghost_piece_);
  }
}

void BlocksGridController::UpdateTime(float delta_time)
{
  grid_.Update(delta_time);
}

bool BlocksGridController::Animating() const
{
 return grid_.Animating();
}

void BlocksGridController::Reset()
{
  grid_.Reset();

  grid_.ClearPieceRegion(ghost_piece_);
  piece_factory_.Reset();
  PickNewCurrentPiece();
}

void BlocksGridController::PickNewCurrentPiece()
{
  current_piece_seed_ = next_piece_seed_;
  int random_seed = piece_factory_.GetRandomSeed();

  next_piece_grid_.ClearPieceRegion(next_piece_);

  current_piece_ = piece_factory_.GenerateRandomPiece(START, next_piece_seed_);
  next_piece_ = piece_factory_.GenerateRandomPiece(NP_START, random_seed);
  next_piece_seed_ = random_seed;

  next_piece_grid_.ClearPieceRegion(next_piece_);
  next_piece_grid_.UpdatePieceRegion(next_piece_);

  grid_.AddNewPiece(current_piece_);
  next_piece_grid_.AddNewPiece(next_piece_);

  DeepCopyCurrentPieceToGhostPiece();

  ghost_piece_active_ = true;
  can_swap_           = true;
}

void BlocksGridController::Update()
{
  if (Animating() && current_piece_)
  {
    current_piece_ = nullptr;
  }
  else if (!grid_.CanPieceMoveDirection(current_piece_, DOWN) && !grid_.Lost())
  {
    if (grid_.RowsNeedClearing())
      return;

    // FIXME Find a better way to determine if we have lost or not...
    for (auto const& p : current_piece_->GetPoints())
    {
      if (p.y() >= START.y())
      {
        grid_.Lose();
        return;
      }
    }

    PickNewCurrentPiece();
  }
  else
  {
    MoveCurrentPieceDown();

    if (IsCurrentPieceNearGhostPiece())
    {
      ghost_piece_active_ = false;
      grid_.ClearPieceRegion(ghost_piece_);
    }
    else
    {
      ghost_piece_active_ = true;
    }
  }
}

bool BlocksGridController::IsCurrentPieceNearGhostPiece() const
{
  int lowest  = std::numeric_limits<int>::max();
  for (auto const& p : current_piece_->GetPoints())
    if (p.y() < lowest)
      lowest = p.y();

  int highest = 0;
  for (auto const& p : ghost_piece_->GetPoints())
    if (p.y() >= highest)
      highest = p.y();

  if (ghost_piece_active_ && lowest - highest <= 1)
    return true;

  return false;
}

void BlocksGridController::Draw(SDLHandler* sdl_handler)
{
  grid_.Draw(sdl_handler);
  next_piece_grid_.Draw(sdl_handler);

  if (grid_.Lost())
  {
    grid_.Lose();
  }
}

void BlocksGridController::MoveCurrentPieceRight()
{
  MovePiece(RIGHT);
}

void BlocksGridController::MoveCurrentPieceLeft()
{
  MovePiece(LEFT);
}

void BlocksGridController::MoveCurrentPieceDown()
{
  MovePiece(DOWN);
}

void BlocksGridController::HardDropCurrentPiece()
{
  while (grid_.CanPieceMoveDirection(current_piece_, DOWN))
    MovePiece(DOWN);

  Update();
}

void BlocksGridController::MovePiece(Direction direction)
{
  if (grid_.Lost() || Animating())
    return;

  if (grid_.CanPieceMoveDirection(current_piece_, direction))
  {
    grid_.ClearPieceRegion(current_piece_);

    current_piece_->MoveDirection(direction);
    UpdateGhostPiece();

    grid_.UpdatePieceRegion(current_piece_);
  }
}

void BlocksGridController::RotateCurrentPieceClockwise()
{
  if (grid_.Lost() || Animating())
    return;

  if (grid_.CanPieceRotateClockwise(current_piece_))
  {
    grid_.ClearPieceRegion(current_piece_);

    current_piece_->RotateClockwise();
    UpdateGhostPiece();

    grid_.UpdatePieceRegion(current_piece_);
  }
}

void BlocksGridController::SwapNextPiece()
{
  if (can_swap_ && grid_.CanPieceMoveDirection(current_piece_, DOWN))
  {
    next_piece_grid_.ClearPieceRegion(next_piece_);
    grid_.ClearPieceRegion(current_piece_);

    grid_.RemovePiece(current_piece_);

    next_piece_    = piece_factory_.GenerateRandomPiece(NP_START, current_piece_seed_);
    current_piece_ = piece_factory_.GenerateRandomPiece(START, next_piece_seed_);

    grid_.AddNewPiece(current_piece_);

    int temp = current_piece_seed_;
    current_piece_seed_ = next_piece_seed_;
    next_piece_seed_    = temp;

    next_piece_grid_.UpdatePieceRegion(next_piece_);
    grid_.UpdatePieceRegion(current_piece_);

    UpdateGhostPiece();

    can_swap_ = false;
  }
}

}
