// -*- Mode: C++; indent-tabs-mode: nil; tab-width: 2 -*-
/*
 * Copyright (C) 2014 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Brandon Schaefer <brandontschaefer@gmail.com>
 */

#include "Colors.h"
#include "Cell.h"

#include "SDLHandler.h"

#include <stdio.h>

namespace blocks
{

namespace
{
  Color const DEFAULT_COLOR = BLACK;
}

float const BORDER = 3.0f;

Cell::Cell(int x, int y, int size)
  : x_((float)x)
  , y_((float)y)
  , size_((float)size)
  , color_(DEFAULT_COLOR)
  , wall_(false)
  , phased_(true)
{
}

void Cell::SetColor(Color const& color)
{
  if (!wall_)
    color_ = color;
}

void Cell::SetWall()
{
  phased_ = false;
  wall_   = true;
  color_  = DARK_GRAY;
}

void Cell::RemoveWall()
{
  wall_   = false;
  phased_ = true;
  color_  = DEFAULT_COLOR;
}

void Cell::SetPhased()
{
  phased_ = true;
}

void Cell::UnsetPhased()
{
  phased_ = false;
}

bool Cell::Marked() const
{
  if (phased_)
    return false;

  return true;
}

// Maps an x/y positions + size to a vertice based on w/h of the viewport, then normalize it
constexpr float m_to_v(float xy, float wh, float size = 0.0f) { return (xy - wh/2.0f + size) / (wh/2.0f); }

/*|-----------|
 *|  3    4-0 |
 *|     /     |
 *| 5-1    2  |
 *|-----------|
*/

void Cell::DrawLightBorder(float w, float h)
{
  float four_zero_x = m_to_v(x_, w, size_);
  float four_zero_y = m_to_v(y_, h, size_);

  float five_one_x  = m_to_v(x_, w);
  float five_one_y  = m_to_v(y_, h);

  float three_x     = m_to_v(x_, w);
  float three_y     = m_to_v(y_, h, size_);

  GLfloat cell_vertices[] = { five_one_x , five_one_y , 0.0f,
                              three_x    , three_y    , 0.0f,
                              four_zero_x, four_zero_y, 0.0f };

  glEnableVertexAttribArray(0);

  Color lighter_color = color_;
  lighter_color.color_[0] *= 1.8f;
  lighter_color.color_[1] *= 1.8f;
  lighter_color.color_[2] *= 1.8f;

  glVertexAttrib4fv(1, lighter_color.color_);

  glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, cell_vertices);
  glDrawArrays(GL_TRIANGLES, 0, 3);
}

void Cell::DrawDarkBorder(float w, float h)
{
  float four_zero_x = m_to_v(x_, w, size_);
  float four_zero_y = m_to_v(y_, h, size_);

  float five_one_x  = m_to_v(x_, w);
  float five_one_y  = m_to_v(y_, h);

  float two_x       = m_to_v(x_, w, size_);
  float two_y       = m_to_v(y_, h);

  GLfloat cell_vertices[] = { five_one_x , five_one_y , 0.0f,
                              two_x      , two_y      , 0.0f,
                              four_zero_x, four_zero_y, 0.0f };

  glEnableVertexAttribArray(0);

  Color darker_color = color_;
  darker_color.color_[0] *= 0.6f;
  darker_color.color_[1] *= 0.6f;
  darker_color.color_[2] *= 0.6f;

  glVertexAttrib4fv(1, darker_color.color_);

  glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, cell_vertices);
  glDrawArrays(GL_TRIANGLES, 0, 3);
}

void Cell::DrawCell(float w, float h)
{
  float four_zero_x = m_to_v(x_, w, size_ - BORDER);
  float four_zero_y = m_to_v(y_, h, size_ - BORDER);

  float five_one_x  = m_to_v(x_ + BORDER, w);
  float five_one_y  = m_to_v(y_ + BORDER, h);

  float two_x       = m_to_v(x_, w, size_ - BORDER);
  float two_y       = m_to_v(y_ + BORDER, h);

  float three_x     = m_to_v(x_ + BORDER, w);
  float three_y     = m_to_v(y_, h, size_ - BORDER);

  GLfloat cell_vertices[] = { five_one_x , five_one_y , 0.0f,
                              two_x      , two_y      , 0.0f,
                              three_x    , three_y    , 0.0f,
                              four_zero_x, four_zero_y, 0.0f };

  glEnableVertexAttribArray(0);
  glVertexAttrib4fv(1, color_.color_);

  glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, cell_vertices);
  glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
}

// FIXME switch over to use a VBO, things get slow when you add more cells
void Cell::Draw(SDLHandler* sdl_handler)
{
  GLint viewport[4];
  glGetIntegerv(GL_VIEWPORT, viewport);
  float w = (float)viewport[2];
  float h = (float)viewport[3];

  DrawLightBorder(w, h);
  DrawDarkBorder(w, h);
  DrawCell(w, h);
}

}
