// -*- Mode: C++; indent-tabs-mode: nil; tab-width: 2 -*-
/*
 * Copyright (C) 2014 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Brandon Schaefer <brandontschaefer@gmail.com>
 */

#include "Grid.h"

using namespace std;

namespace blocks
{

constexpr int CalcGridSize(int wh, int size, int spacing) { return (wh * size + (wh - 1) * spacing); }

Grid::Grid(Point const& start, Point const& xy_cells, int cell_size, int spacing)
  : x_(start.x())
  , y_(start.y())
  , x_cells_(xy_cells.x())
  , y_cells_(xy_cells.y())
  , cell_size_(cell_size)
  , spacing_(spacing)
{
  InitGrid();
}

void Grid::InitGrid()
{
  int count = 0;
  int h = 0;

  int x_grid_size = CalcGridSize(x_cells_, cell_size_, spacing_);
  int y_grid_size = CalcGridSize(y_cells_, cell_size_, spacing_);

  for (int i = x_; i < x_ + x_grid_size; i += cell_size_ + spacing_)
  {
    grid_.push_back(vector<Cell>());
    h = 0;

    for (int j = y_; j < y_ + y_grid_size; j += cell_size_ + spacing_)
    {
      grid_[count].push_back(Cell(i,j, cell_size_));

      if (count <= 0)
      {
        grid_[count][h].SetWall();
      }
      else if (count >= x_cells_ - 1)
      {
        grid_[count][h].SetWall();
      }

      h++;
    }

    grid_[count][0].SetWall();
    grid_[count][y_cells_-1].SetWall();
    count++;
  }
}

int Grid::GetXCells() const
{
  return x_cells_;
}

int Grid::GetYCells() const
{
  return y_cells_;
}


bool Grid::InBounds(Point const& p) const
{
  if (p.x() < 0 || p.x() >= x_cells_ ||
      p.y() < 0 || p.y() >= y_cells_)
  {
    return false;
  }

  return true;
}

bool Grid::IsMarked(Point const& p) const
{
  if (InBounds(p) && grid_[p.x()][p.y()].Marked())
    return true;

  return false;
}

void Grid::SetPhased(Point const& point)
{
  if (InBounds(point))
    grid_[point.x()][point.y()].SetPhased();
}

void Grid::UnsetPhased(Point const& point)
{
  if (InBounds(point))
    grid_[point.x()][point.y()].UnsetPhased();
}

void Grid::SetColor(Point const& point, Color const& color)
{
  if (InBounds(point))
    grid_[point.x()][point.y()].SetColor(color);
}

void Grid::Draw(SDLHandler* sdl_handler)
{
  for (auto i = 0; i < (int)grid_.size(); i++)
  {
    for (auto j = 0; j < (int)grid_[0].size(); j++)
    {
      grid_[i][j].Draw(sdl_handler);
    }
  }
}

}
