// -*- Mode: C++; indent-tabs-mode: nil; tab-width: 2 -*-
/*
 * Copyright (C) 2014 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Brandon Schaefer <brandontschaefer@gmail.com>
 */

#include "Background.h"
#include "Cell.h"
#include "Colors.h"
#include "MainLoop.h"
#include "Shaders.h"

#include <stdlib.h>
#include <ctime>

#include <iostream>

using namespace std;

namespace blocks
{

int const SCREEN_X_CELLS = 16;
int const SCREEN_Y_CELLS = 30;
int const SPACING = 1;

int const BARRIER = 39;
int const Y_BARRIER = 150;

int const RP_SIZE = 75;

float const START_SPEED = 0.5f;
float const DURATION    = 30.0f;

MainLoop::MainLoop()
  : sdl_handler_("Blocks")
  , screen_(sdl_handler_.Width(), sdl_handler_.Height(), SCREEN_X_CELLS, SCREEN_Y_CELLS, SPACING)
  , grid_controller_(screen_, sdl_handler_.Width(), sdl_handler_.Height())
  , piece_speed_(DURATION, START_SPEED)
  , pause_button_(sdl_handler_.Width()  - RP_SIZE, sdl_handler_.Height() - RP_SIZE, RP_SIZE)
  , reset_button_(0, sdl_handler_.Height() - RP_SIZE, RP_SIZE)
  , current_speed_(START_SPEED)
  , paused_(true)
  , mouse_moved_(false)
{
  sdl_handler_.update.connect      (sigc::mem_fun(this, &MainLoop::Update));
  sdl_handler_.draw.connect        (sigc::mem_fun(this, &MainLoop::Draw));
  sdl_handler_.motion_event.connect(sigc::mem_fun(this, &MainLoop::MotionEvent));
  sdl_handler_.mouse_event.connect (sigc::mem_fun(this, &MainLoop::MouseEvent));
  sdl_handler_.key_event.connect   (sigc::mem_fun(this, &MainLoop::KeyEvent));

  piece_speed_.update.connect(sigc::mem_fun(this, &MainLoop::UpdatePieceSpeed));
  piece_speed_.done.connect  (sigc::mem_fun(this, &MainLoop::DonePieceSpeed));
  piece_speed_.Start();

  pause_button_.SetColor(BLUE.color_);
  reset_button_.SetColor(RED.color_);

  glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
}

MainLoop::~MainLoop()
{
}

void MainLoop::Update(SDLHandler* sdl_handler, float delta_time)
{
  if (!paused_)
  {
    piece_speed_.Tick(delta_time);
    grid_controller_.UpdateTime(delta_time);
  }
}

void MainLoop::UpdatePieceSpeed()
{
  if (!grid_controller_.Animating())
  {
    grid_controller_.Update();
  }
}

void MainLoop::DonePieceSpeed()
{
  if (current_speed_ > 0.11f)
  {
    printf("[MainLoop] Reducing Speed from: %1.4f to %1.4f\n", current_speed_, current_speed_ - 0.025f);
    current_speed_ -= 0.025f;
  }

  piece_speed_.Restart(DURATION, current_speed_);
}

void MainLoop::Draw(SDLHandler* sdl_handler)
{
  glClear(GL_COLOR_BUFFER_BIT);

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glViewport(0, 0, sdl_handler_.Width(), sdl_handler_.Height());
  glUseProgram(sdl_handler->GetGLProgram());

  reset_button_.Draw(sdl_handler);

  GLfloat vVertices[] = { 1.0f,  1.0f, 0.0f,
                          1.0f, -1.0f, 0.0f,
                         -1.0f, -1.0f, 0.0f,
                         -1.0f,  1.0f, 0.0f,
                          1.0f,  1.0f, 0.0f,
                         -1.0f, -1.0f, 0.0f };

  grid_controller_.Draw(sdl_handler);

  if (paused_)
  {

    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, vVertices);
    glEnableVertexAttribArray(0);

    Color color = BLACK;
    color.color_[3] = 0.5f;

    glVertexAttrib4fv(1, color.color_);

    glDrawArrays(GL_TRIANGLES, 0, 6);
  }

  pause_button_.Draw(sdl_handler);
}

void MainLoop::KeyEvent(SDLHandler* sdl_handler, unsigned char key, int keysym, int b)
{
  // Keys that pause shouldn't stop
  switch (keysym)
  {
    case XK_p:
      paused_ = !paused_;
      break;
    case SDLK_F11:
      sdl_handler_.ToggleFullscreen();
      break;
    case SDLK_ESCAPE:
      sdl_handler_.Exit();
      break;
    default:
      break;
  }

  if (paused_)
    return;

  // Keys that pause should stop
  switch (keysym)
  {
    case SDLK_RIGHT:
      grid_controller_.MoveCurrentPieceRight();
      break;
    case SDLK_LEFT:
      grid_controller_.MoveCurrentPieceLeft();
      break;
    case SDLK_UP:
      grid_controller_.RotateCurrentPieceClockwise();
      break;
    case SDLK_DOWN:
      grid_controller_.MoveCurrentPieceDown();
      break;
    case SDLK_SPACE:
      grid_controller_.HardDropCurrentPiece();
      break;
    case SDLK_LSHIFT:
      grid_controller_.SwapNextPiece();
      break;
    case SDLK_r:
      ResetBlocks();
      break;
    default:
      break;
  }
}

void MainLoop::ResetBlocks()
{
  grid_controller_.Reset();
  current_speed_ = START_SPEED;
  piece_speed_.Restart(DURATION, current_speed_);
}

void MainLoop::MouseEvent(SDLHandler* sdl_handler, int x, int y, unsigned int button, bool down)
{
  // OpenGL Y different then my gird coord system
  int y_rev = sdl_handler_.Height() - y;

  if (!down && pause_button_.MouseInside(x, y_rev))
  {
    paused_ = !paused_;
    return;
  }

  if (paused_)
    return;

  static int down_x = 0;
  static int down_y = 0;

  if (!down && reset_button_.MouseInside(x, y_rev))
  {
    ResetBlocks();
    return;
  }

  if (y > 200)
  {
    if (down)
    {
      down_x = x;
      down_y = y;
    }
    else
    {
      if (std::abs(down_x - x) < 15 && std::abs(down_y - y) < 15)
        grid_controller_.HardDropCurrentPiece();
    }
  }

  mouse_moved_ = false;
}

void MainLoop::MotionEvent(SDLHandler* sdl_handler, int x, int y, int dx, int dy, int timestamp)
{
  if (paused_)
    return;

  static int last_time = 0;
  static int d_x = 0;
  static int d_y = 0;

  mouse_moved_ = true;

  if (timestamp - 200 > last_time)
  {
      last_time = timestamp;
      return;
  }

  last_time = timestamp;
  d_x += dx;
  d_y += dy;

  // Hackish, so things dont go crazy when dragging with touch event
  if (std::abs(d_x) >= BARRIER)
  {
      if (d_x > 0)
      {
        grid_controller_.MoveCurrentPieceRight();
      }
      else
      {
        grid_controller_.MoveCurrentPieceLeft();
      }

      d_x = 0;
      d_y = 0;
  }

  if (std::abs(d_y) >= BARRIER)
  {
      if (d_y > 0)
      {
        grid_controller_.MoveCurrentPieceDown();
        d_y = 0;
        d_x = 0;
      }
      else if (std::abs(d_y) > Y_BARRIER)
      {
        grid_controller_.RotateCurrentPieceClockwise();
        d_y = 0;
        d_x = 0;
      }
  }
}

void MainLoop::Start()
{
  sdl_handler_.StartMainLoop();
}

}

int main()
{
  srand((unsigned)time(nullptr));

  blocks::MainLoop ml;
  ml.Start();
}
