// -*- Mode: C++; indent-tabs-mode: nil; tab-width: 2 -*-
/*
 * Copyright (C) 2014 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Brandon Schaefer <brandontschaefer@gmail.com>
 */

#include "Piece.h"
#include "Colors.h"

#include <algorithm>

namespace blocks
{

Piece::Piece()
  : color_(BLACK)
  , is_ghost_(false)
{
}

Piece::Piece(Piece& piece)
{
  piece.color_ = color_;
  piece.pivot_ = pivot_;

  for (auto const& p : points_)
    piece.points_.push_back(p);
}

std::vector<Point> Piece::GetPoints() const
{
  return points_;
}

Color Piece::GetColor() const
{
  return color_;
}

void Piece::SetAsGhostPiece()
{
  color_ = LIGHT_GRAY;
  is_ghost_ = true;
}

bool Piece::IsGhostPiece() const
{
  return is_ghost_;
}

void Piece::RemovePoint(Point const& point)
{
  points_.erase(std::remove(points_.begin(), points_.end(), point), points_.end());
}

bool Piece::PointInPiece(Point const& point) const
{
  for (auto p : points_)
  {
    if (point == p)
      return true;
  }

  return false;
}

void Piece::MoveDirection(Direction const& direction)
{
  for (auto& p : points_)
    p = p.MoveDirection(direction);

  pivot_ = pivot_.MoveDirection(direction);
}

Direction Piece::OppositeDirection(Direction const& direction) const
{
  return Direction((direction + 2) % 4);
}

void Piece::MoveOppositeDirection(Direction const& direction)
{
  Direction opp = OppositeDirection(direction);
  MoveDirection(opp);
}

// {A, B} -> {B, -A}
Point RotatePoint270(Point const& point)
{
  return {point.y(), -point.x()};
}

void Piece::RotateClockwise()
{
  for (auto& p : points_)
  {
    Point tmp = p - pivot_;
    tmp = RotatePoint270(tmp);
    p = tmp + pivot_;
  }
}

// {A, B} -> {-B, A}
Point RotatePoint90(Point const& point)
{
  return {-point.y(), point.x()};
}

void Piece::UndoRotation()
{
  for (auto& p : points_)
  {
    Point tmp = p - pivot_;
    tmp = RotatePoint90(tmp);
    p = tmp + pivot_;
  }
}

void Piece::MoveDownPointsAboveRow(int row)
{
  for (auto& p : points_)
  {
    if (p.y() >= row)
    {
      p = p.MoveDirection(DOWN);
    }
  }
}

}
