// -*- Mode: C++; indent-tabs-mode: nil; tab-width: 2 -*-
/*
 * Copyright (C) 2014 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Brandon Schaefer <brandontschaefer@gmail.com>
 */

#include "SDLHandler.h"
#include "Shaders.h"

#include <sys/time.h>

#include <iostream>
#include <string.h>

#include <unistd.h>

using namespace std;


SDLHandler::SDLHandler(char const* title)
  : width_(0)
  , height_(0)
  , fullscreen_(false)
  , done_(false)
{
  if (SDL_Init(SDL_INIT_VIDEO) < 0)
  {
    printf("FAILED sdl init\n");
    exit(0);
  }

  SDL_DisplayMode mode;
  SDL_GetDisplayMode(0, 0, &mode);

  if (mode.w <= 0 || mode.h <= 0)
  {
    printf("Error getting screen size: width(%i) height(%i)\n", mode.w, mode.h);
    exit(0);
  }

  width_ = mode.w;
  height_ = mode.h;

  // Remove 5% of the edge of the screen
  width_  -= (int)(width_ * 0.05);
  height_ -= (int)(height_ * 0.05);

  window_ = SDL_CreateWindow(title, SDL_WINDOWPOS_UNDEFINED,
                             SDL_WINDOWPOS_UNDEFINED, width_, height_,
                             SDL_WINDOW_SHOWN);

  SDL_GetWindowSize(window_, &width_, &height_);

  if (!window_)
  {
    printf("window fail\n");
    exit(0);
  }

  renderer_ = SDL_CreateRenderer(window_, -1, SDL_RENDERER_ACCELERATED);

  if (!renderer_)
  {
    printf("renderer fialed\n");
    exit(0);
  }

  context_ = SDL_GL_CreateContext(window_);

  // add some error checking

  SDL_GL_SetSwapInterval(1);

  printf("%p %p\n", (void*)window_, (void*)renderer_);
  printf("%s....\n", glGetString(GL_SHADING_LANGUAGE_VERSION));

  if ((program_ = shade_loader_.LoadShaders()) == 0)
  {
    printf("Failed to load shaders");
    exit(0);
  }

  SDL_GL_MakeCurrent(window_, context_);

}

SDLHandler::~SDLHandler()
{
}

void SDLHandler::Exit()
{
  SDL_GL_DeleteContext(context_);
  SDL_DestroyRenderer(renderer_);
  SDL_DestroyWindow(window_);
  SDL_Quit();
  done_ = true;
}

int SDLHandler::Width() const
{
  return width_;
}

int SDLHandler::Height() const
{
  return height_;
}

GLint SDLHandler::GetGLProgram() const
{
    return program_;
}

void SDLHandler::ToggleFullscreen()
{
  fullscreen_ = !fullscreen_;
  SDL_SetWindowFullscreen(window_, fullscreen_);
}

void SDLHandler::StartMainLoop()
{
  struct timeval time1, time2;
  struct timezone time_end;
  float delta_time = 0.0f;
  float totaltime = 0.0f;
  int frames = 0;

  gettimeofday(&time1, &time_end);

  while (!done_)
  {
    gettimeofday(&time2, &time_end);
    delta_time = (float)((double)time2.tv_sec - (double)time1.tv_sec + ((double)time2.tv_usec - (double)time1.tv_usec) * 1e-6);
    time1 = time2;

    SDL_Event event;
    while (SDL_PollEvent(&event))
    {
      if (event.type == SDL_FINGERDOWN || event.type == SDL_FINGERUP)
      {
        bool down = false;
        if (event.type == SDL_FINGERDOWN)
          down = true;

        mouse_event.emit(this, (int)(event.tfinger.x * (float)width_), (int)(event.tfinger.y * (float)height_), 0, down);
      }
      else if (event.type == SDL_FINGERMOTION)
      {
        motion_event.emit(this, (int)(event.tfinger.x  * (float)width_),  (int)(event.tfinger.y * (float)height_),
                                (int)(event.tfinger.dx * (float)width_), (int)(event.tfinger.dy * (float)height_),
                                event.motion.timestamp);
      }
      else if (event.type == SDL_KEYDOWN)
      {
        key_event.emit(this, '\0', (int)event.key.keysym.sym, 0);
      }
      else if (event.type == SDL_MOUSEBUTTONDOWN)
      {
        mouse_event.emit(this, (int)event.button.x, (int)event.button.y, 0, true);
      }
      else if (event.type == SDL_MOUSEBUTTONUP)
      {
        mouse_event.emit(this, (int)event.button.x, (int)event.button.y, 0, false);
      }
      else if (event.type == SDL_MOUSEMOTION)
      {
        //motion_event.emit(this, event.motion.x, event.motion.y, event.motion.xrel, event.motion.yrel, event.motion.timestamp);
      }
      else if (event.type == SDL_QUIT)
      {
        done_ = true;
      }
    }

    update.emit(this, delta_time);
    draw.emit(this);
    SDL_GL_SwapWindow(window_);

    totaltime += delta_time;
    frames++;
    if (totaltime >  2.0f)
    {
      printf("%4d frames rendered in %1.4f seconds -> FPS=%3.4f\n", frames, totaltime, (float)frames/totaltime);
      totaltime -= 2.0f;
      frames = 0;
    }
  }
}
